#pragma OPENCL EXTENSION cl_khr_fp64 : enable
__kernel void bfgs( __global double *K,__global double *gs,
  __global double *gs1, __global double *gs2,__global double *gt,
  int n,int m)
{

  int rowgroup=get_group_id(1);
  int colgroup=get_group_id(0);
  int i=get_local_id(1);
  int j=get_local_id(0);
  __local double s[64];
  __local double s1[64];
  __local double s2[64];
  __local double s2t[64];
  __local double t[64];
  __local double tt[64];

  int ioffset=rowgroup*m;
  int joffset=colgroup*m;

  // move parts of the vectors from global to 
  // local memory for fast access
 
  s[j]=gs[j+joffset];
  s1[i]=gs1[i+ioffset];
  s2[j]=gs2[j+joffset];
  s2t[i]=gs2[i+ioffset];
  t[i]=gt[i+ioffset];
  tt[j]=gt[j+joffset];
 
  barrier(CLK_LOCAL_MEM_FENCE);
  
  double tmp=s1[i]*s[j]-t[i]*s2[j]-s2t[i]*tt[j];

  //if (i<j)   // for now do the entire nxn matrix
  K[(i+ioffset)*n+j+joffset]+=tmp;
}

__kernel void bfgs2( __global double *K,__global double *gs,
    __global double *gt,
    __global double *gy,
    __global double *gsy,
    __global double *gyt,
  int n,int m)
{

  int rowgroup=get_group_id(1);
  int colgroup=get_group_id(0);
  int i=get_local_id(1);
  int j=get_local_id(0);
  __local double s[64];
  __local double s1[64];
  __local double s2[64];
  __local double s2t[64];
  __local double t[64];
  __local double tt[64];

  int ioffset=rowgroup*m;
  int joffset=colgroup*m;

  // move parts of the vectors from global to 
  // local memory for fast access
 
  double tmp1=gs[j+joffset];
  double tmp2=gs[i+ioffset];
  double sy=*gsy;
  double a=(sy+*gyt)/(sy*sy);
  s[j]=tmp1;
  s1[i]=a*tmp2;
  s2[j]=tmp1/sy;
  s2t[i]=tmp2/sy;
  t[i]=gt[i+ioffset];
  tt[j]=gt[j+joffset];
 
  barrier(CLK_LOCAL_MEM_FENCE);
  
  double tmp=s1[i]*s[j]-t[i]*s2[j]-s2t[i]*tt[j];

  //if (i<j)   // for now do the entire nxn matrix
  K[(i+ioffset)*n+j+joffset]+=tmp;
}

__kernel void MatVecMul(const __global double* M,
                        const __global double* V,
                        __global double* W,
                        int n,int m)
{
  int row=get_global_id(0);
  int lid=get_local_id(1);
  __local double psum[256];
 
  double tmp=0.0;

  for (int i=lid;i<n;i+=m)
    tmp+=M[row*n+i]*V[i];

  psum[lid]=tmp;
  barrier(CLK_LOCAL_MEM_FENCE);
 
  for (int m1=m/2;m1>0;m1>>=1)   // m>>=1 same as  m1/=2
  {
    if (lid<m1)
    {
      psum[lid]+=psum[lid+m1];
    }
    barrier(CLK_LOCAL_MEM_FENCE);
  }
  
  if (lid==0)
    W[row]=psum[0];
}



__kernel void Vecdot(const __global double* V,
                     const __global double* W,
                      __global double * dot_prod,
                        int n,int m)
{
 
  int lid=get_local_id(0);
  __local double psum[256];
 
  double tmp=0.0;

  for (int i=lid;i<n;i+=m)
    tmp+=V[i]*W[i];

  psum[lid]=tmp;
  barrier(CLK_LOCAL_MEM_FENCE);
 
  for (int m1=m/2;m1>0;m1/=2)
  {
    if (lid<m1)
    {
      psum[lid]+=psum[lid+m1];
    }
    barrier(CLK_LOCAL_MEM_FENCE);
  }
  if (lid==0)
    *dot_prod=psum[0];
 
}


__kernel void Vecnorm(const __global double* V,
                        __global double* norm,
                        int n,int m)
{
  int lid=get_local_id(0);
  __local double psum[256];
 
  double tmp=0.0;

  for (int i=lid;i<n;i+=m)
    tmp+=V[i]*V[i];

  psum[lid]=tmp;
  barrier(CLK_LOCAL_MEM_FENCE);
 
  for (int m1=m/2;m1>0;m1/=2)
  {
    if (lid<m1)
    {
      psum[lid]+=psum[lid+m1];
    }
    barrier(CLK_LOCAL_MEM_FENCE);
  }
  
  *norm=sqrt(psum[0]);
}





/*

__kernel void MatVecMul(const __global double* M,
                        const __global double* gV,
                        __global double* W,
                        int n,int m)
{
  int rowgroup=get_group_id(1);
  int colgroup=get_group_id(0);
  int i=get_local_id(1);
  int j=get_local_id(0);
  __local double V[64];
  __local double tmp[256];

  int ioffset=rowgroup*m;
  int joffset=colgroup*m;

  // move parts of the vectors from global to 
  // local memory for fast access
 
  V[j]=gV[j+joffset];
  
  barrier(CLK_LOCAL_MEM_FENCE);
  int i16=i*16;
  tmp[i16+j]=M[(i+ioffset)*n+j+joffset]*V[j];
  barrier(CLK_LOCAL_MEM_FENCE);
  if (j<8 )
  {
    tmp[i16+j]+=tmp[i16+j+8];
  }
  barrier(CLK_LOCAL_MEM_FENCE);
  if (j<4 )
  {
    tmp[i16+j]+=tmp[i16+j+4];
  }
  barrier(CLK_LOCAL_MEM_FENCE);
  if (j<2 )
  {
    tmp[i16+j]+=tmp[i16+j+2];
  }
  barrier(CLK_LOCAL_MEM_FENCE);
  
  if (j<1 )
  {
    tmp[i16+j]+=tmp[i16+j+1];
  }
  barrier(CLK_LOCAL_MEM_FENCE);
  if (j==0)
    W[i+ioffset]+=tmp[i16+j];
  barrier(CLK_LOCAL_MEM_FENCE);
}
*/

/*  
#define WARP_SIZE 256
__kernel void MatVecMul(const __global double* M,
                        const __global double* V,
                        uint width, uint height,
                        __global double* W)
{
   __local double partialDotProduct[1024];
   // Each work-group computes multiple elements of W
 
   for (uint y = get_group_id(0); y < height; y += get_num_groups(0)) 
   {
      const __global double* row = M + y * width;

      // Each work-item accumulates as many products as necessary
      // into local variable "sum"
      double sum = 0;
      for (uint x = get_local_id(0); x < width; x += get_local_size(0))
         sum += row[x] * V[x];

      // Each partial dot product is stored in shared memory
      partialDotProduct[get_local_id(0)] = sum;

      // Perform parallel reduction to add each work-item's
      // partial dot product together

      // Synchronize to make sure each work-item is done writing to
      // partialDotProduct
      barrier(CLK_LOCAL_MEM_FENCE);

      // Thread local ID within a warp
      uint id = get_local_id(0) & (WARP_SIZE - 1); 

      // Each warp reduces 64 consecutive elements
      double warpResult = 0.0f;
      if (get_local_id(0) < get_local_size(0)/2 )
      {
          volatile __local double* p = partialDotProduct + 2 * get_local_id(0) - id;
          p[0] += p[32];
          p[0] += p[16];
          p[0] += p[8];
          p[0] += p[4];
          p[0] += p[2];
          p[0] += p[1];
          warpResult = p[0];
      }

      // Synchronize to make sure each warp is done reading
      // partialDotProduct before it is overwritten in the next step
      barrier(CLK_LOCAL_MEM_FENCE);

      // The first thread of each warp stores the result of the reduction
      // at the beginning of partialDotProduct
      if (id == 0)
         partialDotProduct[get_local_id(0) / WARP_SIZE] = warpResult;

      // Synchronize to make sure each warp is done writing to
      // partialDotProduct before it is read in the next step
      barrier(CLK_LOCAL_MEM_FENCE);

      // Number of remaining elements after the first reduction
      uint size = get_local_size(0) / (2 * WARP_SIZE);

      // get_local_size(0) is less or equal to 512 on NVIDIA GPUs, so
      // only a single warp is needed for the following last reduction
      // step
      if (get_local_id(0) < size / 2) {
         __local double* p = partialDotProduct + get_local_id(0);
         if (size >= 8)
            p[0] += p[4];
         if (size >= 4)
            p[0] += p[2];
         if (size >= 2)
            p[0] += p[1];
      }

      // Write the result of the reduction to global memory
      if (get_local_id(0) == 0)
         W[y] = partialDotProduct[0];

      // Synchronize to make sure the first work-item is done with
      // reading partialDotProduct
      barrier(CLK_LOCAL_MEM_FENCE);
   }
   
}
*/
