#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif
#define MAX_SOURCE_SIZE (0x100000)

#include <admodel.h>
int * star(ivector& v)
{
  return &(v(v.indexmin()));
}

double * star(dvector& v)
{
  return &(v(v.indexmin()));
}

main()
{
   int n=4096+2048+2048+2048;
   int nn=n*n;
   int m=16;
   int i;
   dmatrix H(1,n,1,n);  //  H is the Hessian (sort of)
   dmatrix K(1,n,1,n);  // K is H inv
   dmatrix K1(1,n,1,n);  // K is H inv
   random_number_generator rng(101);
   dvector s(1,n);
   dvector y(1,n);
   y.fill_randn(rng);
   s.fill_randn(rng);
   // If you want to use a real invserse for comparison
   // uncomment this. But it takes a long time for large matrices
   //H.fill_randn(rng);
   //H=H*trans(H);
   //K=inv(H);
   for (i=1;i<=n;i++)
   {
     K(i,i)=1.0;
     if ((i+1)<=n)
     {
       K(i+1,i)=0.2;
       K(i,i+1)=0.2;
     }
   } 
   K1=K;
  adtimer adt;
  adt.get_elapsed_time_and_reset();
   dvector t=K*y;   //  matrix * vector
   dvector tsave=y*K;   //  matrix * vector
   double sy=s*y;  // vector dot product
   double yt=y*t;  // vector dot product
   dvector s1=((sy + yt)/square(sy))*s;
   dvector s2=s/sy;
   // BFPS update of K = H inv
   for (i=1;i<=n;i++)
   {
     for (int j=1;j<=i;j++)
     {
       K1(i,j)+=s1(i)*s(j)-t(i)*s2(j)-s2(i)*t(j);
       K1(j,i)=K1(i,j);
     }
   } 
   cerr << " CPU time  " << adt.get_elapsed_time_and_reset()/1000. 
        << " seconds" << endl;
   //cout << norm2(K1-inv(H)) << endl;
   // this is the update for the inverse of the Hessian
   //K=K+outer_prod(s1,s)-outer_prod(t,s2)-outer_prod(s2,t);
   // K(i,j)+=s1(i)*s(j)-t(i)*s2(j)-s2(i)*t(j)
   // this is the update for the Hessian
   //H=H+outer_prod(y,y)/(y*s) -outer_prod(H*s,H*s)/(s*(H*s));
   // check that H=inv(H)
   //cout << norm2(K-inv(H)) << endl;
   // now try the GPU
   
   dvector Kstack(1,nn);
   dvector Kstack1(1,nn);

   int ioffset=0;
   for (i=1;i<=n;i++)
   {
     Kstack(1+ioffset,n+ioffset).shift(1)=K(i);
     ioffset+=n;
   }
   //Kstack.initialize();


 // Load the kernel source code into the array source_str
  FILE *fp;
  char *source_str;
  size_t source_size;

  fp = fopen("bfg2.cl", "r");
  if (!fp) {
      fprintf(stderr, "Failed to load kernel source.\n");
      exit(1);
  }
  source_str = (char*)malloc(MAX_SOURCE_SIZE);
  source_size = fread( source_str, 1, MAX_SOURCE_SIZE, fp);
  fclose( fp );

  // Get platform and device information
  cl_platform_id platform_id = NULL;
  cl_device_id device_id = NULL;   
  cl_uint ret_num_devices;
  cl_uint ret_num_platforms;
  cl_int ret = clGetPlatformIDs(1, &platform_id, &ret_num_platforms);
  //ret = clGetDeviceIDs( platform_id, CL_DEVICE_TYPE_ALL, 1, 
  ret = clGetDeviceIDs( platform_id, CL_DEVICE_TYPE_CPU, 1, 
          &device_id, &ret_num_devices);

  // Create an OpenCL context
  cl_context context = clCreateContext( NULL, 1, &device_id, NULL, NULL, &ret);

  // Create a command queue
  cl_command_queue command_queue = clCreateCommandQueue(context, device_id, 0, &ret);

  // Create a program from the kernel source
  cl_program program = clCreateProgramWithSource(context, 1, 
          (const char **)&source_str, (const size_t *)&source_size, &ret);
  cout << "Program creation code = " << ret << endl;

  // Build the program
  ret = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
  cout << "Program build code = " << ret << endl;

  // Create the OpenCL kernel
  cl_kernel kernel2 = clCreateKernel(program, "Vecdot", &ret);
  cout << "Create Kernel2 error code = " << ret << endl;

  // Create the OpenCL kernel
  cl_kernel kernel = clCreateKernel(program, "bfgs", &ret);
  cout << "Create Kernel error code = " << ret << endl;

  // Create the OpenCL kernel
  cl_kernel kernel1 = clCreateKernel(program, "MatVecMul", &ret);
  cout << "Create Kernel kernel1 error code = " << ret << endl;
  // Create memory buffers on the device for each vector 
  cl_mem K_mem_obj = clCreateBuffer(context, CL_MEM_READ_WRITE, 
          nn * sizeof(double), NULL, &ret);
  cl_mem s_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY,
          n * sizeof(double), NULL, &ret);
  cl_mem s1_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY,
          n * sizeof(double), NULL, &ret);
  cl_mem s2_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY,
          n * sizeof(double), NULL, &ret);
  cl_mem t_mem_obj = clCreateBuffer(context, CL_MEM_READ_ONLY,
          n * sizeof(double), NULL, &ret);
  cl_mem y_mem_obj = clCreateBuffer(context, CL_MEM_READ_WRITE, 
          n * sizeof(double), NULL, &ret);
  cl_mem ky_mem_obj = clCreateBuffer(context, CL_MEM_READ_WRITE, 
          n * sizeof(double), NULL, &ret);
  cl_mem ty_mem_obj = clCreateBuffer(context, CL_MEM_READ_WRITE, 
          sizeof(double), NULL, &ret);
  cl_mem sy_mem_obj = clCreateBuffer(context, CL_MEM_READ_WRITE, 
          sizeof(double), NULL, &ret);

  // Copy the lists A and B to their respective memory buffers
  ret = clEnqueueWriteBuffer(command_queue, K_mem_obj, CL_TRUE, 0,
          nn * sizeof(double), star(Kstack), 0, NULL, NULL);
  adt.get_elapsed_time_and_reset();
  ret = clEnqueueWriteBuffer(command_queue, s_mem_obj, CL_TRUE, 0, 
          n * sizeof(double), star(s), 0, NULL, NULL);
  ret = clEnqueueWriteBuffer(command_queue,s1_mem_obj,CL_TRUE,0, 
          n * sizeof(double), star(s1), 0, NULL, NULL);
  ret = clEnqueueWriteBuffer(command_queue,s2_mem_obj,CL_TRUE,0, 
          n * sizeof(double), star(s2), 0, NULL, NULL);
  ret = clEnqueueWriteBuffer(command_queue,t_mem_obj,CL_TRUE,0, 
          n * sizeof(double), star(t), 0, NULL, NULL);
  ret = clEnqueueWriteBuffer(command_queue,y_mem_obj,CL_TRUE,0, 
          n * sizeof(double), star(y), 0, NULL, NULL);
  dvector ky(1,n);
  ky.initialize();
  ret = clEnqueueWriteBuffer(command_queue,ky_mem_obj,CL_TRUE,0, 
          n * sizeof(double), star(ky), 0, NULL, NULL);

 int m1=256;
 size_t global_item_size1[3]={n,m1,1};  // nxn work items
 size_t local_item_size1[3] = {1,m1,1}; // Process in groups of size mxm 

  int no=0;
  ret = clSetKernelArg(kernel1,no++, sizeof(cl_mem), (void *)&K_mem_obj);
  ret = clSetKernelArg(kernel1,no++, sizeof(cl_mem), (void *)&y_mem_obj);
  ret = clSetKernelArg(kernel1,no++, sizeof(cl_mem), (void *)&ky_mem_obj);
  ret = clSetKernelArg(kernel1,no++, sizeof(int), (void *)&n);
  ret = clSetKernelArg(kernel1,no++, sizeof(int), (void *)&m);
  // Execute the OpenCL kernel on the list
 ret = clEnqueueNDRangeKernel(command_queue, kernel1, 2, NULL,
                global_item_size1, local_item_size1, 0, NULL, NULL);
 if (ret!=CL_SUCCESS)
      cout << "kernel1 execution failed" << endl;

  no=0;
  // dot product of t=K*y and y to be put in ty
  ret = clSetKernelArg(kernel1,no++, sizeof(cl_mem), (void *)&ky_mem_obj);
  ret = clSetKernelArg(kernel1,no++, sizeof(cl_mem), (void *)&y_mem_obj);
  ret = clSetKernelArg(kernel1,no++, sizeof(cl_mem), (void *)&ty_mem_obj);
  ret = clSetKernelArg(kernel1,no++, sizeof(int), (void *)&n);
  ret = clSetKernelArg(kernel1,no++, sizeof(int), (void *)&m);
 size_t global_item_size2[3]={n,1,1};  // nxn work items
 size_t local_item_size2[3] = {m1,1,1}; // Process in groups of size mxm 

 ret = clEnqueueNDRangeKernel(command_queue, kernel2, 1, NULL,
                global_item_size2, local_item_size2, 0, NULL, NULL);

  no=0;
  ret = clSetKernelArg(kernel,no++, sizeof(cl_mem), (void *)&K_mem_obj);
  ret = clSetKernelArg(kernel,no++, sizeof(cl_mem), (void *)&s_mem_obj);
  ret = clSetKernelArg(kernel,no++, sizeof(cl_mem), (void *)&s1_mem_obj);
  ret = clSetKernelArg(kernel,no++, sizeof(cl_mem), (void *)&s2_mem_obj);
  ret = clSetKernelArg(kernel,no++, sizeof(cl_mem), (void *)&ky_mem_obj);
  ret = clSetKernelArg(kernel,no++, sizeof(int), (void *)&n);
  ret = clSetKernelArg(kernel,no++, sizeof(int), (void *)&m);

 // Execute the OpenCL kernel on the list

  size_t global_item_size[3]={n,n,1};  // nxn work items
  size_t local_item_size[3] = {m,m,1}; // Process in groups of size mxm 
 ret = clEnqueueNDRangeKernel(command_queue, kernel, 2, NULL,
                global_item_size, local_item_size, 0, NULL, NULL);
 if (ret!=CL_SUCCESS)
      cout << "kernel execution failed" << endl;
   ioffset=0;
   dmatrix Kreturn(1,n,1,n);
   for (i=1;i<=n;i++)
   {
     Kreturn(i)=Kstack(1+ioffset,n+ioffset).shift(1);
     ioffset+=n;
   }
   //cout << setw(10) << Kreturn << endl << endl;
  cerr << " GPU time  " << adt.get_elapsed_time_and_reset()/1000. 
       << " seconds " << endl;

  ret = clEnqueueReadBuffer(command_queue, K_mem_obj, CL_TRUE, 0,
            nn * sizeof(double), star(Kstack1), 0, NULL, NULL);
 if (ret!=CL_SUCCESS)
      cout << "K read  failed" << endl;

   ioffset=0;
   for (i=1;i<=n;i++)
   {
     Kreturn(i)=Kstack1(1+ioffset,n+ioffset).shift(1);
     ioffset+=n;
   }

   cout << "Norm of difference between CPU and GPU BFGS update of K = H inv" 
        << endl << norm2(K1-Kreturn) << " for an " << n << " times "
        << n << " matrix " << endl;

}


